__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]


__title__ = "CADETMatch"
__summary__ = "CADETMatch is a parameter estimation and error modeling library for CADET"
__author__ = "William Heymann"
__email__ = "w.heymann@fz-juelich.de"
__license__ = "GNU General Public License v3 (GPLv3)"
__copyright__ = "2020 %s" % __author__
__version__ = "0.6.23"
__uri__ = "https://github.com/modsim/CADET-Match"
